/*
 * kgraft_patch_bsc1014271
 *
 * Fix for CVE-2016-9576, bsc#1014271
 *
 *  Upstream commit:
 *  a0ac402cfcdc ("Don't feed anything but regular iovec's to blk_rq_map_user_iov")
 *
 *  SLE12-SP2 commit:
 *  fc1fe0e3bad405be09494c5df2e17c0ab38eaf66
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/blkdev.h>
#include <linux/uio.h>

static int (*kgr__blk_rq_unmap_user)(struct bio *bio);
static void (*kgr_blk_rq_bio_prep)(struct request_queue *q, struct request *rq, struct bio *bio);
static struct bio *(*kgr_bio_copy_user_iov)(struct request_queue *q, struct rq_map_data *map_data, const struct iov_iter *iter, gfp_t gfp_mask);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "__blk_rq_unmap_user", (void *)&kgr__blk_rq_unmap_user },
	{ "blk_rq_bio_prep", (void *)&kgr_blk_rq_bio_prep },
	{ "bio_copy_user_iov", (void *)&kgr_bio_copy_user_iov },
};

/* inlined */
static bool kgr_iovec_gap_to_prv(struct request_queue *q,
			     struct iovec *prv, struct iovec *cur)
{
	unsigned long prev_end;

	if (!queue_virt_boundary(q))
		return false;

	if (prv->iov_base == NULL && prv->iov_len == 0)
		/* prv is not set - don't check */
		return false;

	prev_end = (unsigned long)(prv->iov_base + prv->iov_len);

	return (((unsigned long)cur->iov_base & queue_virt_boundary(q)) ||
		prev_end & queue_virt_boundary(q));
}

/* patched function */
int kgr_blk_rq_map_user_iov(struct request_queue *q, struct request *rq,
			struct rq_map_data *map_data,
			const struct iov_iter *iter, gfp_t gfp_mask)
{
	struct bio *bio;
	int unaligned = 0;
	struct iov_iter i;
	struct iovec iov, prv = {.iov_base = NULL, .iov_len = 0};

	if (!iter || !iter->count)
		return -EINVAL;

	/*
	 * Fix CVE-2016-9576
	 *  +2 lines
	 */
	if (!iter_is_iovec((struct iov_iter *) iter))
		return -EINVAL;

	iov_for_each(iov, i, *iter) {
		unsigned long uaddr = (unsigned long) iov.iov_base;

		if (!iov.iov_len)
			return -EINVAL;

		/*
		 * Keep going so we check length of all segments
		 */
		if ((uaddr & queue_dma_alignment(q)) ||
		    kgr_iovec_gap_to_prv(q, &prv, &iov))
			unaligned = 1;

		prv.iov_base = iov.iov_base;
		prv.iov_len = iov.iov_len;
	}

	if (unaligned || (q->dma_pad_mask & iter->count) || map_data)
		bio = kgr_bio_copy_user_iov(q, map_data, iter, gfp_mask);
	else
		bio = bio_map_user_iov(q, iter, gfp_mask);

	if (IS_ERR(bio))
		return PTR_ERR(bio);

	if (map_data && map_data->null_mapped)
		bio_set_flag(bio, BIO_NULL_MAPPED);

	if (bio->bi_iter.bi_size != iter->count) {
		/*
		 * Grab an extra reference to this bio, as bio_unmap_user()
		 * expects to be able to drop it twice as it happens on the
		 * normal IO completion path
		 */
		bio_get(bio);
		bio_endio(bio);
		kgr__blk_rq_unmap_user(bio);
		return -EINVAL;
	}

	if (!bio_flagged(bio, BIO_USER_MAPPED))
		rq->cmd_flags |= REQ_COPY_USER;

	blk_queue_bounce(q, &bio);
	bio_get(bio);
	kgr_blk_rq_bio_prep(q, rq, bio);
	return 0;
}

static int kgr_patch_bsc1014271_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int kgr_patch_bsc1014271_init(void)
{
	return kgr_patch_bsc1014271_kallsyms();
}
